/*-
 * Copyright (C)2019 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2019 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: xml2pg.c,v 1.7 2019/05/11 20:23:06 babolo Exp $"

#define MIFE_COMPAT 5

#include <sys/types.h>
#include <sys/event.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <sys/time.h>
#include <sysexits.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <getCGIparm.h>
#include <mife.h>

#define XML2PG_TYNUM 21
#define XML2PG_ATNUM 10
#define XML2PG_FLIST 124

typedef struct {
    BLIN_flag  flags;
#       define XML2PG_DUMP    0x000100
#       define XML2PG_DUUP /* 0x000200 */ (XML2PG_DUMP << 1)
#       define XML2PG_FULL    0x000400
#       define XML2PG_FUUL /* 0x000800 */ (XML2PG_FULL << 1)
#       define XML2PG_COPY    0x001000
#       define XML2PG_COPZ /* 0x002000 */ (XML2PG_COPY << 1)
#       define XML2PG_SHFT    0x020000
#       define XML2PG_NAME    0x040000
#       define XML2PG_BLNK    0x080000
    BLIN_flag  subflags;
    xmlDoc    *xdoc;
} xml2pg_cf;

int xml2pg_dump __P((xml2pg_cf *options, xmlNode *xobj, int ex, int exx));

static char *xtnm[XML2PG_TYNUM + 1] =
{ "none"               /*  0 */
, "ELEMENT_NODE"       /*  1 */
, "ATTRIBUTE_NODE"     /*  2 */
, "TEXT_NODE"          /*  3 */
, "CDATA_SECTION_NODE" /*  4 */
, "ENTITY_REF_NODE"    /*  5 */
, "ENTITY_NODE"        /*  6 */
, "PI_NODE"            /*  7 */
, "COMMENT_NODE"       /*  8 */
, "DOCUMENT_NODE"      /*  9 */
, "DOCUMENT_TYPE_NODE" /* 10 */
, "DOCUMENT_FRAG_NODE" /* 11 */
, "NOTATION_NODE"      /* 12 */
, "HTML_DOCUMENT_NODE" /* 13 */
, "DTD_NODE"           /* 14 */
, "ELEMENT_DECL"       /* 15 */
, "ATTRIBUTE_DECL"     /* 16 */
, "ENTITY_DECL"        /* 17 */
, "NAMESPACE_DECL"     /* 18 */
, "XINCLUDE_START"     /* 19 */
, "XINCLUDE_END"       /* 20 */
, "DOCB_DOCUMENT_NODE" /* 21 */
};

static char *atnm[XML2PG_ATNUM + 1] =
{ "none"        /*  0 */
, "CDATA"       /*  1 */
, "ID"          /*  2 */
, "IDREF"       /*  3 */
, "IDREFS"      /*  4 */
, "ENTITY"      /*  5 */
, "ENTITIES"    /*  6 */
, "NMTOKEN"     /*  7 */
, "NMTOKENS"    /*  8 */
, "ENUMERATION" /*  9 */
, "NOTATION"    /* 10 */
};

static char *ftnm[XML_WITH_LZMA + 1] =
{ "none"       /*  0 */
, "THREAD"     /*  1 */
, "TREE"       /*  2 */
, "OUTPUT"     /*  3 */
, "PUSH"       /*  4 */
, "READER"     /*  5 */
, "PATTERN"    /*  6 */
, "WRITER"     /*  7 */
, "SAX1"       /*  8 */
, "FTP"        /*  9 */
, "HTTP"       /* 10 */
, "VALID"      /* 11 */
, "HTML"       /* 12 */
, "LEGACY"     /* 13 */
, "C14N"       /* 14 */
, "CATALOG"    /* 15 */
, "XPATH"      /* 16 */
, "XPTR"       /* 17 */
, "XINCLUDE"   /* 18 */
, "ICONV"      /* 19 */
, "ISO8859X"   /* 20 */
, "UNICODE"    /* 21 */
, "REGEXP"     /* 22 */
, "AUTOMATA"   /* 23 */
, "EXPR"       /* 24 */
, "SCHEMAS"    /* 25 */
, "SCHEMATRON" /* 26 */
, "MODULES"    /* 27 */
, "DEBUG"      /* 28 */
, "DEBUG_MEM"  /* 29 */
, "DEBUG_RUN"  /* 30 */
, "ZLIB"       /* 31 */
, "ICU"        /* 32 */
, "LZMA"       /* 33 */
};

static u_char *
forcopy(xml2pg_cf *options, const u_char *n) {
#   define blin_internal_flags (options->flags)
    int     ex = EX_OK;
    u_char *m;
    char   *u;

    if  (!n) {
        if  (!(m = (u_char*)strdup("\\N"))) ifBLIN_QW0("strdup \\N");
    } else if (!(m = (u_char*)(u = strdup((char*)n)))) {
        ifBLIN_QW0("strdup %s", n);
    } else if ((ex = getCGIparmpass3(0, (u_char*)"copy", &m))) {
        ifBLIN_QW0("getCGIparmpass3 %s", m);
        if  ((u_char*)u != m) free(u);
        free(m);
        m = NULL;
    } else {
        if  ((u_char*)u != m) free(u);
    }
    return(m);
#   undef blin_internal_flags
}

static int
xobjdump(xml2pg_cf *options, xmlNode *xobj) {
#   define blin_internal_flags (options->flags)
    int          ex = EX_OK;
    xmlBufferPtr buffer;
    int          l = 0;
    u_char      *n0;

    if  (options->flags & XML2PG_DUMP) {
        if  (options->flags & XML2PG_DUUP) l = 1;
        if  (!(buffer = xmlBufferCreate())) {
            ifBLIN_QW0("xmlBufferCreate");
            ex = EX_OSERR;
            goto out;
        }
        if  (0 > xmlNodeDump(buffer, options->xdoc, xobj, l, l)) {
            ifBLIN_QW0("xmlNodeDump");
            ex = EX_OSERR;
            goto out;
        }
        if  (!(n0 = forcopy(options, xmlBufferContent(buffer)))) {
            ifBLIN_QW0("forcopy n0");
            ex = EX_OSERR;
            goto out;
        }
        printf("\t%s", n0);
    }
out:
    return(ex);
#   undef blin_internal_flags
}

static int
nsout(xml2pg_cf *options, xmlNs *ns, int ex, int exx, int shift) {
#   define blin_internal_flags (options->flags)
    u_char  *n1 = NULL;
    u_char  *n0;

    if  (~options->flags & XML2PG_SHFT) shift = 0;
    if  (ns) {
        if  (ns->context) {
            exx = xml2pg_dump(options, xmlDocGetRootElement(ns->context), ex, ++exx);
            if  (0 > exx) {
                ifBLIN_QX0("xml2pg_dump");
                ex = exx;
                goto out;
        }   }
        if  (!(n0 = forcopy(options, ns->prefix))) {
            ifBLIN_QW0("forcopy n0");
            ex = -EX_OSERR;
            goto out;
        }
        if  (!(n1 = forcopy(options, ns->href))) {
            ifBLIN_QW0("forcopy n1");
            ex = -EX_OSERR;
            goto out;
        }
        printf("%d\t%d", ++exx, ex);
        if  (options->flags & XML2PG_NAME) {
            printf("\t%s", xtnm[((ns->type < XML2PG_TYNUM) ? ns->type : 0)]);
        } else {
            printf("\t%d", ((ns->type < XML2PG_TYNUM) ? ns->type : 0) + shift);
        }
        printf("\t%s\t%s", n0, n1);
        if  (options->flags & XML2PG_COPZ) printf("\t\\N");
        if  (!!xobjdump(options, (xmlNode *)ns)) {
            ifBLIN_QW0("xobjdump ns");
            ex = -EX_OSERR;
            goto out;
        }
        printf("\n");
        free(n1);
        free(n0);
    }
out:
    return(exx);
#   undef blin_internal_flags
}

int
xml2pg_dump( xml2pg_cf    *options
           , xmlNode      *xobj
           , int           ex   /*        */
           , int           exx  /*                   */
           ) {
#   define blin_internal_flags (options->flags)
    int      nex;

    ifBLIN_QX2( "+%08X %u<-%u\n"
              , options->flags
              , ex
              , exx
              )
    ;
    for (; xobj; xobj = xobj->next) {
        nex = exx;
        ifBLIN_QX3("%i %i %i", ex, exx, xobj->type);
        switch(xobj->type) {
        case XML_ELEMENT_NODE:
        case XML_ATTRIBUTE_NODE:
        case XML_TEXT_NODE:
        case XML_CDATA_SECTION_NODE:
        case XML_ENTITY_REF_NODE:
        case XML_ENTITY_NODE:
        case XML_PI_NODE:
        case XML_COMMENT_NODE:
        case XML_DOCUMENT_NODE:
        case XML_DOCUMENT_TYPE_NODE:
        case XML_DOCUMENT_FRAG_NODE:
        case XML_NOTATION_NODE:
        case XML_HTML_DOCUMENT_NODE:
        case XML_DTD_NODE:
        case XML_ELEMENT_DECL:
        case XML_ATTRIBUTE_DECL:
        case XML_ENTITY_DECL:
        case XML_NAMESPACE_DECL:
        case XML_XINCLUDE_START:
        case XML_XINCLUDE_END:
        case XML_DOCB_DOCUMENT_NODE:
            {   xmlChar *cur = NULL;
                u_char  *n0 = NULL;
                u_char  *n1;
                char    *n2;

                if  (!(n1 = forcopy(options, xobj->content))) {
                    ifBLIN_QW0("forcopy n1");
                    ex = -EX_OSERR;
                    goto out;
                }
                if  ((options->flags & XML2PG_FULL) && !!xobj->ns && !!xobj->ns->prefix) {
                    asprintf(&n2, "%s:%s", xobj->ns->prefix, xobj->name);
                } else if (!(n2 = strdup((const char *)xobj->name))) {
                    ifBLIN_QW0("strdup n2");
                    ex = -EX_OSERR;
                    goto out;
                }
                printf("%d\t%d", exx, ex);
                if  (options->flags & XML2PG_NAME) {
                    printf( "\t%s", xtnm[(xobj->type < XML2PG_TYNUM) ? xobj->type : 0]);
                } else {
                    printf( "\t%d", (xobj->type < XML2PG_TYNUM) ? xobj->type : 0);
                }
                printf("\t%s\t%s", n2, n1);
                if  (options->flags & XML2PG_COPZ) {
                    cur = xmlNodeListGetString(options->xdoc, xobj, 1);
                    if  (!(n0 = forcopy(options, cur))) {
                        ifBLIN_QW0("forcopy n0");
                        ex = -EX_OSERR;
                        goto out;
                    }
                    printf("\t%s", n0);
                    xmlFree(cur);
                    free(n0);
                }
                if  (!!xobjdump(options, xobj)) {
                    ifBLIN_QW0("xobjdump xobj");
                    ex = -EX_OSERR;
                    goto out;
                }
                printf("\n");
                free(n2);
                free(n1);
                for (xmlAttr *attr = xobj->properties; attr; attr = attr->next) {
                    xmlChar *val = xmlNodeListGetRawString(xobj->doc, attr->children, 1);
                    u_char  *n0;
                    char    *n2;

                    if  (!(n0 = forcopy(options, val))) {
                        ifBLIN_QW0("forcopy n0");
                        ex = -EX_OSERR;
                        goto out;
                    }
                    if  ((options->flags & XML2PG_FUUL) && !!attr->ns && !!attr->ns->prefix) {
                        asprintf(&n2, "%s:%s", attr->ns->prefix, attr->name);
                    } else if (!(n2 = strdup((const char *)attr->name))) {
                        ifBLIN_QW0("strdup n2");
                        ex = -EX_OSERR;
                        goto out;
                    }
                    printf("%d\t%d", ++exx, nex);
                    if  (options->flags & XML2PG_NAME) {
                        printf("\t%s", xtnm[((attr->type < XML2PG_TYNUM) ? attr->type : 0)]);
                    } else {
                        printf( "\t%d"
                              , ((attr->type < XML2PG_TYNUM) ? attr->type : 0)
                              + ((options->flags & XML2PG_SHFT) ? 128 : 0)
                              );
                    }
                    printf("\t%s", n0);
                    if  (options->flags & XML2PG_COPZ) {
                        if  (options->flags & XML2PG_NAME) {
                            printf("\t%s", atnm[(attr->atype < XML2PG_ATNUM) ? attr->atype : 0]);
                        } else {
                            printf("\t%d", (attr->atype < XML2PG_ATNUM) ? attr->atype : 0);
                    }   }
                    if  (!!xobjdump(options, (xmlNode *)attr)) {
                        ifBLIN_QW0("xobjdump attr");
                        ex = -EX_OSERR;
                        goto out;
                    }
                    printf("\n");
                    free(n2);
                    free(n0);
                    free(val);
                    if  (  (~options->flags & XML2PG_FUUL)
                        && (0 > (exx = nsout(options, attr->ns, nex, exx, 32)))
                        ) {
                        ifBLIN_QX0("nsout ns");
                        ex = exx;
                        goto out;
                }   }
           /*
            *   if  (xobj->doc) {
            *       exx = xml2pg_dump(options, xmlDocGetRootElement(xobj->doc), nex, ++exx);
            *       if  (0 > exx) {
            *           ifBLIN_QX0("xml2pg_dump");
            *           ex = exx;
            *           goto out;
            *   }   }
            */
                if  (  (~options->flags & XML2PG_FULL)
                    && (0 > (exx = nsout(options, xobj->ns, nex, exx, 64)))
                    ) {
                    ifBLIN_QX0("nsout ns");
                    ex = exx;
                    goto out;
                }
                if  (0 > (exx = nsout(options, xobj->nsDef, nex, exx, 96))) {
                    ifBLIN_QX0("nsout nsDef");
                    ex = exx;
                    goto out;
                }
                exx = xml2pg_dump(options, xobj->children, nex, ++exx);
            }
            break;
        default: {
                char *t;

                t = (options->flags & XML2PG_NAME) ? "none" : "0";
                printf( (options->flags & XML2PG_COPZ)
                      ? "%d\t%d\t%s\t%s\t\\N\t\\N\n"
                      : "%d\t%d\t%s\t%s\t\\N\n"
                      , ++exx
                      , ex
                      , t
                      , xobj->name
                      );
            }
            break;
    }   }
    ex = exx;
out:
    ifBLIN_QX2("-%08X %u<-%u\n", options->flags, ex, ++exx);
    return(ex);
#   undef blin_internal_flags
}

static void
usage(int ex) {
    const char *fl[XML2PG_FLIST];
    int         tifl;
    int         ifl;
    int         i;

    fprintf( stderr
           , "xml2pg @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "    -b        - do not create blank nodes\n"
             "    -c[c]     - COPY object recursive/plain\n"
             "    -d[d]     - recursive dump node to last column\n"
             "    -f[f]     - name with namespace\n"
             "    -n        - symbolic type names\n"
             "    -q        - quiet\n"
             "    -s        - shift type numbers\n"
             "    -v        - verbose\n"
             " Feature list:\n"
           )
    ;
    if  (0 > (tifl = xmlGetFeaturesList(&ifl, fl))) {
        ifBLIN_QX0("xmlGetFeaturesList");
    } else {
        for (i = 0; i < ifl; ++i) {
            fprintf(stderr, "%s\n", fl[i]);
    }   }
    fprintf(stderr, " Has features:\n");
    for (int i = 1; i <= XML_WITH_LZMA; ++i) {
        fprintf(stderr, "    XML_WITH_%s	%d\n", ftnm[i], xmlHasFeature(i));
    }
    fprintf(stderr, " List of node types:\n");
    for (i = 1; i <= XML2PG_TYNUM; ++i) {
        fprintf(stderr, "    %2d XML_%s\n", i, xtnm[i]);
    }
    fprintf(stderr, " List of attribute types:\n");
    for (i = 1; i <= XML2PG_ATNUM; ++i) {
        fprintf(stderr, "    %2d XML_ATTRIBUTE_%s\n", i, atnm[i]);
    }
    exit(ex);
}

int
main(int argc, char **argv) {
#   define blin_internal_flags (options->flags)
    const char          *flar = "bcdfnqsv=:?";
    xmlNode             *xobj = NULL;
    int                  ex = EX_OK;
    babolo_opts         *bos = NULL;
    xml2pg_cf           *options;
    off_t                xoffset;
    ssize_t              xsize;
    xmlParserCtxtPtr     ctxt;
    int                  xnum;
    mife_descriptor     *mdi;
    const char          *o;
    int                  c;

    LIBXML_TEST_VERSION;
    if  (!(options = calloc(1, sizeof(*options)))) {
        warnx("Malloc failed #1");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    options->flags = BLIN_BIT0;
    if  (!(bos = babolo_openopts(Bpars_NONU | Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("babolo_openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (babolo_setopts(bos, Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("babolo_setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    while (0 < (c = babolo_getopts(bos))) {
        switch (c) {
        case 'b': options->flags |= XML2PG_BLNK;
                ; break;
        case 'c': options->flags |= (options->flags & XML2PG_COPY) << 1;
                ; options->flags |= XML2PG_COPY;
                ; break;
        case 'd': options->flags |= (options->flags & XML2PG_DUMP) << 1;
                ; options->flags |= XML2PG_DUMP;
                ; break;
        case 'f': options->flags |= (options->flags & XML2PG_FULL) << 1;
                ; options->flags |= XML2PG_FULL;
                ; break;
        case 'n': options->flags |= XML2PG_NAME;
                ; break;
        case 'q': options->flags &= ~BLIN_MASK;
                ; break;
        case 's': options->flags |= XML2PG_SHFT;
                ; break;
        case 'v': BLIN_VERBOSE(options->flags);
                ; break;
        case '=': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; switch(o[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(bos);                exit(EX_OK);
                  case '3': babolo_dumpopts(bos);              exit(EX_OK);
                  }
                ; /* FALLTHROUGH */
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c", c);
                ; usage(EX_USAGE);
    }   }
    options->subflags = ( (((BLIN_BIT0 | BLIN_BIT1) & options->flags)
                        | (options->flags >> 3)) & BLIN_MASK
                        ) | MIFE_SLID
    ;
    if  (!(mdi = mife_init(options->subflags))) {
        ifBLIN_QW0("mife_init stdin");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > mife_ctlfdsc(mdi, fileno(stdin))) {
        ifBLIN_QW0("mife_ctlfdsc stdin");
        ex = EX_IOERR;
        goto out;
    }
    xoffset = 0;
    xnum = 0;
    if  (options->flags & XML2PG_BLNK) xmlKeepBlanksDefault(0);
    for (int i = 0; ; ++i, xoffset += xsize) {
        if  (0 > (xsize = mife_read(mdi, 0, xoffset))) {
            ifBLIN_QW0("mife_read stdin");
            ex = EX_IOERR;
            goto out;
        }
        if  (!xsize) {
            ifBLIN_QX0("null stdin");
            ex = EX_DATAERR;
            goto out;
        }
        if  (!(o = mife_get(mdi, xoffset))) {
            ifBLIN_QW0("mife_get stdin");
            ex = EX_IOERR;
            goto out;
        }
        if  (!i) {
            if  (!(ctxt = xmlCreatePushParserCtxt(NULL, NULL, o, xsize, NULL))) {
                ifBLIN_QW0("xmlCreatePushParserCtxt");
                errno = EDOOFUS;
                ex = EX_SOFTWARE;
                goto out;
            }
        } else {
            if  (!!(ex = xmlParseChunk(ctxt, o, xsize, 0))) {
                ifBLIN_QX0("xmlParseChunk %d", ex);                     /* XXXX Enum xmlParserErrors */
                errno = EDOOFUS;
                ex = EX_SOFTWARE;
                goto out;
        }   }
        if  (mdi->flags & MIFE_EOFL) {
            if  (!!(ex = xmlParseChunk(ctxt, o, 0, 0))) {
                ifBLIN_QX0("xmlParseChunk %d", ex);                     /* XXXX Enum xmlParserErrors */
                errno = EDOOFUS;
                ex = EX_SOFTWARE;
                goto out;
            }
            break;
    }   }
    if  (!ctxt->wellFormed) {
        ifBLIN_QW0("Failed to parse");
        errno = EFTYPE;
        ex = EX_SOFTWARE;
        goto out;
    }
    options->xdoc = ctxt->myDoc;
    xmlFreeParserCtxt(ctxt);
    xobj = xmlDocGetRootElement(options->xdoc);
    if  (options->flags & XML2PG_COPY) {
        if  (0 > (xnum = xml2pg_dump(options, xobj, 0, ++xnum))) {
            ifBLIN_QX0("xml2pg_dump");
            ex = -xnum;
            goto out;
        } else {
            ex = 0;
    }   }
    xmlFreeDoc(options->xdoc);
    xmlCleanupParser();
    xmlMemoryDump();
out:
    exit(ex);
#   undef blin_internal_flags
}
